open Mltype
open Printf
open Projet

(****************************************************************************************************)

let (etiquette, newreg, set_etiquette) =
  let compteur = ref 0 and etiq = ref "" in
  (fun () -> incr compteur;
    if !etiq = "" then
      sprintf "__%d" !compteur
    else
      sprintf "__%s_%d" !etiq !compteur),
  (fun () -> incr compteur; !compteur),
  (fun str -> etiq := str)
;;

(****************************************************************************************************)

exception Not_supported
module Symbol = Map.Make(String)

type binding =
  | Litteral of int
  | Register of int
  | Absolute of string
  | Environment of int

type asm_vinstr =
  | APUTARG of int
  | ABACKUP of binding * int * int
  | ALABEL of string
  | ACALLFUN of binding * int
  | AONEOP of expr * int
  | ABINOP of expr * binding * int
  | AMOV of binding * int
  | AOTHER of string
  | ACALLPRIM of string * int
  | APUSH of binding
  | ACLEANUP of int
  | ACALLSETUP
  | ACALLPOST

(****************************************************************************************************)

let rec bytecode table accu accu2 = function

  | VAR a ->			(* Printf.printf "Recherche de %s...\n" a; *)
     (Symbol.find a table, accu, accu2)

  | FUN (a, b) ->		(* Printf.printf "Definition de fun %s -> ...\n" a; *)
      let etiq1 = etiquette () and reg1 = newreg () and reg2 = newreg () in
      let (accu, new_table) =
	let (_, new_table, c_accu) =
	  let copier x (n, new_table, c_accu) =
	    if x = a then (n, Symbol.add x (Register reg2) new_table, c_accu)
	    else match Symbol.find x table with
	      | (Register _) | (Environment _) as a ->
		  (n+1,
		   Symbol.add x (Environment n) new_table,
		   ABACKUP (a, n, reg1) :: c_accu)
	      | (Litteral _) | (Absolute _) as a ->
		  (n, Symbol.add x a new_table, c_accu)
	  in Ident.fold copier (var_libres b) (1, Symbol.empty, [])
	in
	let n = List.length c_accu in
	let accu = ACALLPOST
		   :: ACLEANUP (n+1)
		   :: ACALLPRIM ("malloc", reg1)
		   :: APUSH (Litteral ((n+1) * 4))
		   :: ACALLSETUP
		   :: accu in
	(c_accu @ accu, new_table)
      in
      let accu2 =
	let (result, accu, accu2) = bytecode new_table [APUTARG reg2; ALABEL etiq1] accu2 b in
	(List.rev accu) :: accu2
      in (Register reg1, accu, accu2)

  | APP (a, b) ->
      let (b, accu, accu2) = bytecode table accu accu2 b in
      let (a, accu, accu2) = bytecode table accu accu2 a in
      let reg1 = newreg () in
      (Register reg1, ACALLFUN (a, reg1) :: AMOV (b, reg1) :: accu, accu2)

  | LET (a, b, c) ->
      let (b, accu, accu2) = bytecode table accu accu2 b in
      bytecode (Symbol.add a b table) accu accu2 c

  | INT a -> (Litteral a, accu, accu2)
  | UNIT () -> (Litteral 0, accu, accu2)

  | (PLUS (a, b)) | (MINUS (a, b)) | (TIMES (a, b)) | (DIV (a, b)) | (REM (a, b))
  | (EQUAL (a, b)) | (LESS (a, b)) | (SET (a, b)) | (PROJ (a, b)) as expr ->
      let (b, accu, accu2) = bytecode table accu accu2 b in
      let (a, accu, accu2) = bytecode table accu accu2 a in
      let reg1 = newreg () in
      let accu = ABINOP (expr, b, reg1) :: AMOV (a, reg1) :: accu in
      (Register reg1, accu, accu2)

  | (OPP a) | (BANG a) | (REF a) as expr ->
      let (a, accu, accu2) = bytecode table accu accu2 a in
      let reg1 = newreg () in
      let accu = AONEOP (expr, reg1) :: AMOV (a, reg1) :: accu in
      (Register reg1, accu, accu2)

  | IF (a, b, c) as expr ->
      let reg1 = newreg () in
      let etiq1 = etiquette () and etiq2 = etiquette () in
      let (a, accu, accu2) = bytecode table accu accu2 a in
      let accu =
	AOTHER ("jz " ^ etiq1)
	:: AONEOP (expr, reg1)
	:: AMOV (a, reg1)
	:: accu in
      let (b, accu, accu2) = bytecode table accu accu2 b in
      let accu =
	ALABEL etiq1
	:: AOTHER ("jmp " ^ etiq2)
	:: AMOV (b, reg1)
	:: accu in
      let (c, accu, accu2) = bytecode table accu accu2 c in
      let accu =
	ALABEL etiq2
	:: AMOV (c, reg1)
	:: accu
      in (Register reg1, accu, accu2)

  | SEQ (a, b) ->
      let (_, accu, accu2) = bytecode table accu accu2 a
      in bytecode table accu accu2 b

  | CALLPRIM (f, liste) ->
      let reg1 = newreg () in
      let accu = ACALLSETUP :: accu in
      let (accu, accu2) =
	let aux (accu, accu2) a =
	  let (a, accu, accu2) = bytecode table accu accu2 a in
	  (APUSH a :: accu, accu2)
	in List.fold_left aux (accu, accu2) (List.rev liste)
      in
      let accu = ACALLPOST
		 :: ACLEANUP (List.length liste)
		 :: ACALLPRIM (f, reg1)
		 :: accu
      in (Register reg1, accu, accu2)

  | TUPLE liste ->
      let reg1 = newreg () in
      let accu = ACALLPOST
		 :: ACLEANUP 1
		 :: ACALLPRIM ("malloc", reg1)
		 :: APUSH (Litteral (List.length liste * 4))
		 :: ACALLSETUP
		 :: accu in
      let (_, accu, accu2) =
	let aux (n, accu, accu2) a =
	  let (a, accu, accu2) = bytecode table accu accu2 a in
	  (n+1, ABACKUP (a, n, reg1) :: accu, accu2)
	in List.fold_left aux (0, accu, accu2) liste
      in (Register reg1, accu, accu2)

  | Op _ -> raise Not_supported
;;

(****************************************************************************************************)

type operand =
  | IMMEDIATE of string
  | DIRECT of string
  | INDIRECT of int * string * string * int

type asm_pinstr =
  | PLABEL of string
  | PNONE of string
  | PSINGLE of string * operand
  | PDOUBLE of string * operand * operand

(****************************************************************************************************)

type registers = NonReg | OneReg of int | TwoReg of int * int

let reg_of = function
  | (ABACKUP (Register reg1, _, reg2))
  | (ACALLFUN (Register reg1, reg2))
  | (ABINOP (_, Register reg1, reg2))
  | (AMOV (Register reg1, reg2)) -> TwoReg (reg1, reg2)
  | (APUTARG reg) | (APUSH (Register reg)) | (ACALLPRIM (_, reg))
  | (ABACKUP (_, _, reg)) | (ACALLFUN (_, reg)) | (AONEOP (_, reg))
  | (ABINOP (_, _, reg)) | (AMOV (_, reg)) -> OneReg reg
  | (AOTHER _) | (ACLEANUP _) | (APUSH _) | ACALLPOST | ACALLSETUP
  | (ALABEL _) -> NonReg

(****************************************************************************************************)

module Int = struct type t = int let compare = Pervasives.compare end
module RegMap = Map.Make(Int)
module RegSet = Set.Make(Int)

let allouer liste =
  let derniers =
    let enregistrer deja instr = match reg_of_instr instr with
      | TwoReg (reg1, reg2) -> RegMap.add reg1 instr (RegMap.add reg2 instr deja)
      | OneReg reg -> RegMap.add reg instr deja
      | NonReg -> deja
    in List.fold_left enregistrer RegMap.empty liste
  in
  let libres = ref RegSet.empty and maximum = ref 0 in
  let alloc_reg table instr i =
    if RegMap.mem i table then
      table
    else
      let nreg =
	try
	  let nreg = RegSet.min_elt !libres in
	  libres := RegSet.remove nreg !libres;
	  if !libres = RegSet.empty then libres := RegSet.add (nreg+1) !libres;
	  nreg
	with Not_found ->
	  libres := RegSet.add 1 !libres; 0
      in
      maximum := max !maximum nreg;
      if instr == RegMap.find i derniers then libres := RegSet.add i !libres;
      RegMap.add i nreg table
  in
  let generer table instr = match reg_of_instr instr with
    | TwoReg (reg1, reg2) -> alloc_reg (alloc_reg table instr reg1) instr reg2
    | OneReg reg -> alloc_reg table instr reg
    | NonReg -> table
  in (!maximum + 1, List.fold_left generer RegMap.empty liste)
;;

(****************************************************************************************************)

let general_registers = [| "%ebx", "%ecx", "%edx", "%esi", "%edi", "%ebp", "%eax" |]

let get = function
  | Litteral i -> sprintf "$%d" i
  | Register i -> if i < 4 then general_registers.(i) else sprintf "%d(%%ebp)" (i-4)
  | Absolute s -> s
  | Environment i -> sprintf "%d(%%edi)" i

let output_instr outchan = function
  | APUTARG i -> fprintf outchan "\tmovl %%eax, %s\n" (get (Register i))
  | ABACKUP (src, n, dst) -> fprintf outchan "\tmovl %s, %d(%s)\n" (get src) n (get (Register i))
  | ALABEL s -> fprintf outchan "%s:\n" s
  | ACALLFUN (src, dst) ->
      let dst = get (Register dst) in
      fprintf outchan "\tpushal\n\tmovl %s, %%eax\n\tmovl %s, %%edi\n" dst (get src);
      fprintf outchan "\tcall *(%%edi)\n\tpopal\n\tmovl %%eax, %s\n" dst
  | AONEOP (OPP _, i) -> fprintf "\toppl %s\n" (get (Register i))
  | AONEOP (BANG _, i) -> let r = get (Register i) in fprintf outchan "\tmovl (%s), %s\n" r r
  | ABINOP (PLUS (_, _), src, dst) -> fprintf outchan "\taddl %s, %s\n" (get src) (get (Register dst))
  | ABINOP (MINUS (_, _), src, dst) -> fprintf outchan "\tsubl %s, %s\n" (get src) (get (Register dst))
  | ABINOP (TIMES (_, _), src, dst) -> fprintf outchan "\timull %s, %s\n" (get src) (get (Register dst))
  | ABINOP (DIV (_, _), src, dst) ->
      let dst = get (Register dst) in
      fprintf outchan "\tpushl %%ebx\n\tpushl %%edx\n\tmovl %s, %%ebx\n\tmovl %s, %%eax\n" (get src) dst;
      fprintf outchan "\tcltd\n\tidivl %%ebx\n\tmovl %%eax, %s\n\tpopl %%edx\n\tpopl %%ebx\n" dst;
  | ABINOP (REM (_, _), src, dst) ->
      let dst = get (Register dst) in
      fprintf outchan "\tpushl %%ebx\n\tpushl %%edx\n\tmovl %s, %%ebx\n\tmovl %s, %%eax\n" (get src) dst;
      fprintf outchan "\tcltd\n\tidivl %%ebx\n\tmovl %%edx, %s\n\tpopl %%edx\n\tpopl %%ebx\n" dst;
  | ABINOP (EQUAL (_, _), src, dst) ->
      let dst = get (Register dst) in
      fprintf outchan "\txorl \tcmlp %s, %s\n"
  | _ -> raise Not_supported
(*
  | ABINOP of expr * binding * int
  | AMOV of binding * int
  | AOTHER of string
  | ACALLPRIM of string * int
  | APUSH of binding
  | ACLEANUP of int
  | ACALLSETUP
  | ACALLPOST
*)

(****************************************************************************************************)

exception Unbound of string

let stdlib accu accu2 = function

  | "afficher_int" ->
      let accu =
	AOTHER "movl %eax, afficher_int"
	:: AOTHER ".comm afficher_int, 4, 4"
	:: AOTHER "leal afficher_int_val, %eax"
	:: AOTHER "movl $__afficher_int, afficher_int_val"
	:: AOTHER ".comm afficher_int_val, 4, 4"
	:: accu in
      let accu2 =
	[ ALABEL "__afficher_int";
	  AOTHER "pushl %edi";
	  AOTHER "pushl $afficher_int_fmt";
	  AOTHER "call printf";
	  ACLEANUP 2;
	  AOTHER "ret";
	  ALABEL "afficher_int_fmt";
	  AOTHER ".string \"%d\\n\"\n" ] :: accu2
      in (accu, accu2)

  | "int_of_string" ->
      let accu =
	AOTHER "movl %eax, int_of_string"
	:: AOTHER ".comm int_of_string, 4, 4"
	:: AOTHER "leal int_of_string_val, %eax"
	:: AOTHER "movl $__int_of_string, int_of_string_val"
	:: AOTHER ".comm int_of_string_val, 4, 4"
	:: accu in
      let accu2 =
	[ ALABEL "__int_of_string";
	  AOTHER "subl $4, %esp";
	  AOTHER "pushl %esp";
	  AOTHER "pushl $int_of_string_fmt";
	  AOTHER "pushl %edi"; 
	  AOTHER "call sscanf";
	  ACLEANUP 3; 
	  AOTHER "popl %edi";
	  AOTHER "ret";
	  ALABEL "int_of_string_fmt";
	  AOTHER ".string \" %d\"" ] :: accu2
      in (accu, accu2)

  | "r" ->
      let accu =
	AOTHER "movl $r_val, r"
	:: AOTHER ".comm r, 4, 4"
	:: accu in
      let accu2 =
	[ ALABEL "r_val";
	  AOTHER ".string \"r\"" ] :: accu2
      in (accu, accu2)

  | "argc" -> (AOTHER "movl %eax, argc" :: AOTHER "movl 4(%esp), %eax" :: AOTHER ".comm argc, 4, 4" :: accu), accu2
  | "argv" -> (AOTHER "movl %eax, argv" :: AOTHER "movl 8(%esp), %eax" :: AOTHER ".comm argv, 4, 4" :: accu), accu2

  | x -> raise (Unbound x)
;;

(****************************************************************************************************)

let rec asm_of_letrec_list table accu accu2 = function
  | [] -> (accu, accu2)
  | (LETREC (f, e)) :: q ->
      set_etiquette f;
      let symbol = etiquette () in
      let table = Symbol.add f (Absolute symbol) table in
      let vlibres = Ident.filter (fun x -> not (Symbol.mem x table)) (var_libres e) in
      let (accu, accu2) =
	let aux x (accu, accu2) = stdlib accu accu2 x in
	Ident.fold aux vlibres (accu, accu2)
      in
      let table = Ident.fold (fun x table -> Symbol.add x (Absolute x) table) vlibres table in
      let (accu, accu2) = asm_of_expr table 0 accu accu2 e in
      let accu =
	AOTHER ("movl %eax, " ^ symbol)
	:: AOTHER (".comm " ^ symbol ^ ", 4, 4")
	:: APOP (Absolute "%eax")
	:: accu
      in asm_of_letrec_list table accu accu2 q
;;

(****************************************************************************************************)

let operand_of_binding = function
  | StackIndex i -> sprintf "%d(%%esp)" (i * 4)
  | HeapIndex i -> sprintf "%d(%%esi)" (i * 4)
  | Argument -> "%edi"
  | Absolute s -> s

let rec output_asm outchan = function
  | APUSH x -> fprintf outchan "\tpushl %s\n" (operand_of_binding x)
  | APOP x -> fprintf outchan "\tpopl %s\n" (operand_of_binding x)
  | AOTHER x -> fprintf outchan "\t%s\n" x
  | ALABEL x -> fprintf outchan "%s:\n" x
  | ACLEANUP x -> fprintf outchan "\taddl $%d, %%esp\n" (x * 4)
;;

(****************************************************************************************************)

let reg_of str =
  if String.contains str '%' then
    String.sub str (String.index str '%' + 1) 3
  else etiquette ()

let rec simplify = function
  | [] -> []
  | APUSH x :: APOP y :: q when x = y -> simplify q
  | APUSH x :: APOP y :: q ->
      let x = operand_of_binding x in
      let y = operand_of_binding y in
      AOTHER (sprintf "movl %s, %s" x y) :: (simplify q)
  | APUSH x :: APUSH y :: APOP z :: APOP t :: q
      when reg_of (operand_of_binding x) <> reg_of (operand_of_binding z) ->
      let x = operand_of_binding x in
      let y = operand_of_binding (match y with
	| StackIndex i -> StackIndex (i-1)
	| y -> y) in
      let z = operand_of_binding z in
      let t = operand_of_binding t in
      let q = simplify q in
      let q = if x <> t then AOTHER (sprintf "movl %s, %s" x t) :: q else q in
      let q = if y <> z then AOTHER (sprintf "movl %s, %s" y z) :: q else q in q
  | a :: q -> a :: (simplify q)
;;

let asm_of_prog prog =
  let (accu, accu2) = asm_of_letrec_list Symbol.empty [] [] prog in
  let accu = simplify (List.rev (AOTHER "ret" :: accu)) in
  let accu2 = List.map simplify accu2 in
  (accu, accu2)
;;

(****************************************************************************************************)

let output_prog outchan prog =
  fprintf outchan ".align 4\n.globl main\n.type main, @function\n\nmain:\n";
  let (accu, accu2) = asm_of_prog prog in
  let aux = List.iter (output_asm outchan) in
  aux accu;
  fprintf outchan "\n";
  List.iter aux accu2 
;;

(****************************************************************************************************)

let output_file fichier progra =
  let fichier = open_out fichier in
  output_prog fichier progra;
  close_out fichier
;;

(****************************************************************************************************)
