open Mltype
open Projet
open Printf

(****************************************************************************************************)

let (etiquette, set_etiquette) =
  let compteur = ref 0 and etiq = ref "" in
  (fun () -> incr compteur;
    if !etiq = "" then
      sprintf "__%d" !compteur
    else
      sprintf "__%s_%d" !etiq !compteur),
  (fun str -> etiq := str)

(****************************************************************************************************)

exception Not_supported
module Environ = Map.Make(String)

type binding =
  | StackIndex of int
  | HeapIndex of int
  | Absolute of string
  | Argument

type asm_instr =
  | APOP of binding
  | APUSH of binding
  | AOTHER of string
  | ALABEL of string
  | ACLEANUP of int

(****************************************************************************************************)

let rec asm_of_expr table level accu accu2 = function

  | VAR a ->			(* Printf.printf "Recherche de %s...\n" a; *)
      begin match Environ.find a table with
	| StackIndex i -> APUSH (StackIndex (level-i-1)) :: accu
	| x -> APUSH x :: accu
      end, accu2

  | FUN (a, b) ->		(* Printf.printf "Definition de fun %s -> ...\n" a; *)
      let etiq1 = etiquette () in
      let (accu, new_table) =			(* copie variables libres *)
	let (_, new_table, c_accu) =
	  let copier x (n, new_table, c_accu) =
	    if x = a then (n, Environ.add x Argument new_table, c_accu)
	    else match Environ.find x table with
	      | (StackIndex _) | (HeapIndex _) | Argument as bind ->
		  let src = match bind with
		    | StackIndex i -> sprintf "%d(%%esp)" (4 * (level-i-1))
		    | HeapIndex  i -> sprintf "%d(%%esi)" (4 * i)
		    | _ -> "%edi"
		  in (n+1,
		      Environ.add x (HeapIndex n) new_table,
		      AOTHER (sprintf "movl %%ebx, %d(%%eax)" (n * 4))
		      :: AOTHER (sprintf "movl %s, %%ebx" src)
		      :: c_accu)
	      | Absolute _ as bind ->
		  (n, Environ.add x bind new_table, c_accu)
	  in Ident.fold copier (var_libres b) (1, Environ.empty, [])
	in
	let accu =
	  APUSH (Absolute "%eax")
	  :: (c_accu @
	      (AOTHER (sprintf "movl $%s, (%%eax)" etiq1)
	       :: ACLEANUP 1
	       :: AOTHER "call malloc"
	       :: AOTHER (sprintf "pushl $%d" ((List.length c_accu + 1) * 4))
	       :: accu))
	in (accu, new_table)
      in
      let accu2 =				(* gnration du code de la fonction *)
	let (accu, accu2) = asm_of_expr new_table 4 [ALABEL etiq1] accu2 b in
	(List.rev (AOTHER "ret" :: APOP Argument :: accu)) :: accu2
      in (accu, accu2)

  | APP (a, b) ->
      let (accu, accu2) = asm_of_expr table level accu accu2 b in
      let accu = APUSH Argument :: APUSH (Absolute "%esi") :: accu in
      let (accu, accu2) = asm_of_expr table (level+3) accu accu2 a in
      let accu =
	APOP (Absolute "%esi")
	:: APOP Argument
	:: AOTHER "movl %edi, 8(%esp)"
	:: AOTHER "call *(%esi)"
	:: AOTHER "movl 8(%esp), %edi"
	:: APOP (Absolute "%esi")
	:: accu
      in (accu, accu2)

  | LET (a, b, c) ->
      let (accu, accu2) = asm_of_expr table level accu accu2 b in
      let table = Environ.add a (StackIndex level) table in
      let (accu, accu2) = asm_of_expr table (level+1) accu accu2 c in
      let accu =
	APUSH (Absolute "%eax")
	:: ACLEANUP 1
	:: APOP (Absolute "%eax")
	:: accu
      in (accu, accu2)

  | INT a -> (APUSH (Absolute (sprintf "$%d" a)) :: accu), accu2
  | UNIT () -> (APUSH (Absolute "$0") :: accu), accu2

  | (PLUS (a, b)) | (MINUS (a, b)) | (TIMES (a, b)) as expr ->
      let operator = match expr with PLUS _ -> "addl" | MINUS _ -> "subl" | _ -> "imull" in
      let (accu, accu2) = asm_of_expr table level accu accu2 b in
      let (accu, accu2) = asm_of_expr table (level+1) accu accu2 a in
      let accu =
	APUSH (Absolute "%eax") 
	:: AOTHER (operator ^ " %ebx, %eax")
	:: APOP (Absolute "%ebx")
	:: APOP (Absolute "%eax")
	:: accu
      in (accu, accu2)

  | OPP a ->
      let (accu, accu2) = asm_of_expr table level accu accu2 a in
      let accu =
	APUSH (Absolute "%eax")
	:: AOTHER "negl %eax"
	:: APOP (Absolute "%eax")
	:: accu
      in (accu, accu2)

  | (DIV (a, b)) | (REM (a, b)) as expr ->
      let result = match expr with DIV _ -> "%eax" | _ -> "%edx" in
      let (accu, accu2) = asm_of_expr table level accu accu2 b in
      let (accu, accu2) = asm_of_expr table (level+1) accu accu2 a in
      let accu =
	APUSH (Absolute result)
	:: AOTHER "idivl %ebx"
	:: AOTHER "cltd"
	:: APOP (Absolute "%ebx")
	:: APOP (Absolute "%eax")
	:: accu
      in (accu, accu2)

  | (EQUAL (a, b)) | (LESS (a, b)) as expr ->
      let operator = match expr with EQUAL _ -> "sete" | _ -> "setl" in
      let (accu, accu2) = asm_of_expr table level accu accu2 b in
      let (accu, accu2) = asm_of_expr table (level+1) accu accu2 a in
      let etiq1 = etiquette () and etiq2 = etiquette () in
      let accu =
	APUSH (Absolute "%ecx")
	:: AOTHER (operator ^ " %cl")
	:: AOTHER "cmpl %ebx, %eax"
	:: AOTHER "xorl %ecx, %ecx"
	:: APOP (Absolute "%ebx")
	:: APOP (Absolute "%eax")
	:: accu
      in (accu, accu2)

  | IF (a, b, c) ->
      let etiq1 = etiquette () and etiq2 = etiquette () in
      let (accu, accu2) = asm_of_expr table level accu accu2 a in
      let accu =
	AOTHER ("jz " ^ etiq1)
	:: AOTHER "orl %eax, %eax"
	:: APOP (Absolute "%eax")
	:: accu in
      let (accu, accu2) = asm_of_expr table level accu accu2 b in
      let accu =
	ALABEL etiq1
	:: AOTHER ("jmp " ^ etiq2)
	:: accu in
      let (accu, accu2) = asm_of_expr table level accu accu2 c in
      (ALABEL etiq2 :: accu, accu2)

  | SEQ (a, b) ->
      let (accu, accu2) = asm_of_expr table level accu accu2 a in
      asm_of_expr table level (APOP (Absolute "%eax") :: accu) accu2 b

  | CALLPRIM (f, liste) ->
      let ((accu, accu2), _) =
	let aux ((accu, accu2), level) x =
	  ((asm_of_expr table level accu accu2 x), (level+1))
	in List.fold_left aux ((accu, accu2), level) (List.rev liste)
      in
      let accu =
	APUSH (Absolute "%eax")
	:: ACLEANUP (List.length liste)
	:: AOTHER ("call " ^ f)
	:: accu
      in (accu, accu2)

  | BANG a ->
      let (accu, accu2) = asm_of_expr table level accu accu2 a in
      let accu =
	APUSH (Absolute "(%eax)")
	:: APOP (Absolute "%eax")
	:: accu
      in (accu, accu2)

  | REF a ->
      let (accu, accu2) = asm_of_expr table level accu accu2 a in
      let accu =
	APUSH (Absolute "%eax")
	:: AOTHER "mov %ebx, (%eax)"
	:: APOP (Absolute "%ebx")
	:: ACLEANUP 1
	:: AOTHER "call malloc"
	:: APUSH (Absolute "$4")
	:: accu
      in (accu, accu2)

  | SET (a, b) ->
      let (accu, accu2) = asm_of_expr table level accu accu2 b in
      let (accu, accu2) = asm_of_expr table (level+1) accu accu2 a in
      let accu =
	APUSH (Absolute "%eax")
	:: AOTHER "movl %ebx, (%eax)"
	:: APOP (Absolute "%ebx")
	:: APOP (Absolute "%eax")
	:: accu
      in (accu, accu2)

  | PROJ (a, b) ->
      let (accu, accu2) = asm_of_expr table level accu accu2 b in
      let (accu, accu2) = asm_of_expr table (level+1) accu accu2 a in
      let accu =
	APUSH (Absolute "%eax")
	:: AOTHER "movl (%eax, %ebx, 4), %eax"
	:: APOP (Absolute "%ebx")
	:: APOP (Absolute "%eax")
	:: accu
      in (accu, accu2)

  | TUPLE liste ->
      let accu =
	ACLEANUP 1
	:: AOTHER "call malloc"
	:: APUSH (Absolute (sprintf "$%d" (List.length liste * 4)))
	:: accu in
      let (_, accu, accu2) =
	let aux (n, accu, accu2) x =
	  let (accu, accu2) = asm_of_expr table level accu accu2 x in
	  let accu =
	    AOTHER (sprintf "movl %%ebx, %d(%%eax)" (n * 4))
	    :: APOP (Absolute "%ebx")
	    :: accu
	  in (n+1, accu, accu2)
	in List.fold_left aux (0, accu, accu2) liste
      in (APUSH (Absolute "%eax") :: accu), accu2

  | Op _ -> raise Not_supported

(****************************************************************************************************)

exception Unbound of string

let stdlib accu accu2 = function

  | "afficher_int" ->
      let accu =
	AOTHER "movl %eax, afficher_int"
	:: AOTHER ".comm afficher_int, 4, 4"
	:: AOTHER "leal afficher_int_val, %eax"
	:: AOTHER "movl $__afficher_int, afficher_int_val"
	:: AOTHER ".comm afficher_int_val, 4, 4"
	:: accu in
      let accu2 =
	[ ALABEL "__afficher_int";
	  AOTHER "pushl %edi";
	  AOTHER "pushl $afficher_int_fmt";
	  AOTHER "call printf";
	  ACLEANUP 2;
	  AOTHER "ret";
	  ALABEL "afficher_int_fmt";
	  AOTHER ".string \"%d\\n\"" ] :: accu2
      in (accu, accu2)

  | "int_of_string" ->
      let accu =
	AOTHER "movl %eax, int_of_string"
	:: AOTHER ".comm int_of_string, 4, 4"
	:: AOTHER "leal int_of_string_val, %eax"
	:: AOTHER "movl $__int_of_string, int_of_string_val"
	:: AOTHER ".comm int_of_string_val, 4, 4"
	:: accu in
      let accu2 =
	[ ALABEL "__int_of_string";
	  AOTHER "subl $4, %esp";
	  AOTHER "pushl %esp";
	  AOTHER "pushl $int_of_string_fmt";
	  AOTHER "pushl %edi"; 
	  AOTHER "call sscanf";
	  ACLEANUP 3; 
	  AOTHER "popl %edi";
	  AOTHER "ret";
	  ALABEL "int_of_string_fmt";
	  AOTHER ".string \" %d\"" ] :: accu2
      in (accu, accu2)

  | "r" ->
      let accu =
	AOTHER "movl $r_val, r"
	:: AOTHER ".comm r, 4, 4"
	:: accu in
      let accu2 =
	[ ALABEL "r_val";
	  AOTHER ".string \"r\"" ] :: accu2
      in (accu, accu2)

  | "argc" -> (AOTHER "movl %eax, argc"
	       :: AOTHER "movl 4(%esp), %eax"
	       :: AOTHER ".comm argc, 4, 4"
	       :: accu), accu2
  | "argv" -> (AOTHER "movl %eax, argv"
	       :: AOTHER "movl 8(%esp), %eax"
	       :: AOTHER ".comm argv, 4, 4"
	       :: accu), accu2

  | x -> raise (Unbound x)

(****************************************************************************************************)

let rec asm_of_letrec_list table accu accu2 = function
  | [] -> (accu, accu2)
  | (LETREC (f, e)) :: q ->
      set_etiquette f;
      let symbol = etiquette () in
      let table = Environ.add f (Absolute symbol) table in
      let vlibres = Ident.filter (fun x -> not (Environ.mem x table)) (var_libres e) in
      let (accu, accu2) =
	let aux x (accu, accu2) = stdlib accu accu2 x in
	Ident.fold aux vlibres (accu, accu2)
      in
      let table = Ident.fold (fun x table -> Environ.add x (Absolute x) table) vlibres table in
      let (accu, accu2) = asm_of_expr table 0 accu accu2 e in
      let accu =
	AOTHER ("movl %eax, " ^ symbol)
	:: AOTHER (".comm " ^ symbol ^ ", 4, 4")
	:: APOP (Absolute "%eax")
	:: accu
      in asm_of_letrec_list table accu accu2 q

(****************************************************************************************************)

let operand_of_binding = function
  | StackIndex i -> sprintf "%d(%%esp)" (i * 4)
  | HeapIndex i -> sprintf "%d(%%esi)" (i * 4)
  | Argument -> "%edi"
  | Absolute s -> s

let rec output_asm outchan = function
  | APUSH x -> fprintf outchan "\tpushl %s\n" (operand_of_binding x)
  | APOP x -> fprintf outchan "\tpopl %s\n" (operand_of_binding x)
  | AOTHER x -> fprintf outchan "\t%s\n" x
  | ALABEL x -> fprintf outchan "%s:\n" x
  | ACLEANUP x -> fprintf outchan "\taddl $%d, %%esp\n" (x * 4)

(****************************************************************************************************)

let reg_of str =
  if String.contains str '%' then
    String.sub str (String.index str '%' + 1) 3
  else etiquette ()

let rec simplify = function
    | [] -> []
    | APUSH x :: APOP y :: q when x = y -> simplify q
    | APUSH x :: APOP y :: q ->
	let x = operand_of_binding x in
	let y = operand_of_binding y in
	AOTHER (sprintf "movl %s, %s" x y) :: (simplify q)
    | APUSH x :: APUSH y :: APOP z :: APOP t :: q
	when reg_of (operand_of_binding x) <> reg_of (operand_of_binding z) ->
	let x = operand_of_binding x in
	let y = operand_of_binding (match y with
				      | StackIndex i -> StackIndex (i-1)
				      | y -> y) in
	let z = operand_of_binding z in
	let t = operand_of_binding t in
	let q = simplify q in
	let q = if x <> t then AOTHER (sprintf "movl %s, %s" x t) :: q else q in
	let q = if y <> z then AOTHER (sprintf "movl %s, %s" y z) :: q else q in q
    | a :: q -> a :: (simplify q)

(****************************************************************************************************)

let asm_of_prog do_simplify prog =
  let (accu, accu2) = asm_of_letrec_list Environ.empty [] [] prog in
  let accu = List.rev (AOTHER "ret" :: AOTHER "xorl %eax, %eax" :: accu) in
  let accu = if do_simplify then simplify accu else accu in
  let accu2 = if do_simplify then List.map simplify accu2 else accu2 in
  (accu, accu2)

let output_prog do_simplify outchan prog =
  fprintf outchan ".align 4\n.globl main\n.type main, @function\n\nmain:\n";
  let (accu, accu2) = asm_of_prog do_simplify prog in
  let aux = List.iter (output_asm outchan) in
  aux accu;
  fprintf outchan "\n";
  List.iter aux accu2

let output_file do_simplify fichier prog =
  let fichier = open_out fichier in
  output_prog do_simplify fichier prog;
  close_out fichier

(****************************************************************************************************)
